             .title        "Namco 108/109/118/119 Reader Plugin"


;05/17/07
;Written by BootGod
             
             ;vectors for standard system calls

send_byte:   .equ 0200h
baton:       .equ 0203h
chk_vram:    .equ 0206h
chk_wram:    .equ 0209h
wr_ppu:      .equ 020ch
read_byte:   .equ 020fh
init_crc:    .equ 0212h
do_crc:      .equ 0215h
finish_crc:  .equ 0218h

crc0:        .equ 0080h
crc1:        .equ 0081h
crc2:        .equ 0082h
crc3:        .equ 0083h

tx:            .equ 00dfh
address:       .equ 00e0h
address_lo:    .equ 00e0h
address_hi:    .equ 00e1h
uaddress:       .equ 00f0h
uaddress_lo:    .equ 00f0h
uaddress_hi:    .equ 00f1h
curbank:    .equ 00e2h
banksleft:    .equ 00e3h
bank0crc:  .equ 00e4h

bank_lo:    .equ 00e8h
bank_hi:    .equ 00e9h

numpages_lo .equ 00eah
numpages_hi .equ 00ebh


temp_crc:    .equ 00ech

;plugin header defines
p_ver: = 1	;plugin version
p_read: = 1	;plugin is for downloading (dumping)
p_write: = 2	;plugin is for uploading (RAM carts / WRAM etc)
p_other: = 3	;plugin is for some other purpose (i.e. BankWatch)
p_var1: = 4	;plugin uses user variable #1
p_var2: = 8	;plugin uses user variable #2
p_var3: = 16	;plugin uses user variable #3
p_var4: = 32	;plugin uses user variable #4

;plugin header data

             .org 0380h
             
             ;plugin description
             .db "Namco 1x8/1x9 carts"
             .fill 03e0h-*,00h

             ;plugin creator(s)
             .db "BootGod"
             .fill 03f8h-*,00h

             ;PRG/CHR/WRAM min/max sizes supported
             .db 32 >> 4,128 >> 4
             .db 16 >> 3,64 >> 3
             .db 0,0

             ;plugin version and flags
             .db p_ver
             .db p_read

;plugin code start here

;check mirroring

             lda #2		;reset mirroring to H (if mapper controlled)
             sta 8000h
             lda #0
             sta 8001h

             ldx #02fh
             
ddloop:      txa
             jsr wr_ppu
             lda #0
             sta 2007h
             dex
             cpx #01fh
             bne ddloop		;load first byte of 2000, 2400, 2800, 2c00 w/ 00h

             ldx #0		;x = H mirroring

             lda #020h
             jsr wr_ppu		;set to PPU $2000

             lda #055h
             sta 2007h		;write value for testing

             lda #028h
             jsr wr_ppu		;set to PPU $2800
             lda 2007h

             lda 2007h
             cmp #055h		;check if value was mirrored here
             beq mirrord	;it was

             inx		;x = V mirroring

             lda #024h
             jsr wr_ppu		;set to PPU $2400
             lda 2007h

             lda 2007h
             cmp #055h		;check if value was mirrored here
             beq mirrord	;it was

             inx		;x = 4 screen
             bne got_mir
             
mirrord:     lda #2		;test for mapper control via CHR A15
             sta 8000h
             lda #80h
             sta 8001h		;try setting to V mirroring

             lda #024h
             jsr wr_ppu		;set to PPU $2400
             lda 2007h

             lda 2007h
             cmp #055h
             bne got_mir	;register did nothing, so is hardwired

             lda #2
             sta 8000h
             lda #0
             sta 8001h		;try setting to back to H mirroring

             lda #028h
             jsr wr_ppu		;set to PPU $2800
             lda 2007h

             lda 2007h
             cmp #055h
             bne got_mir	;register did nothing, so is hardwired

             ldx #04h		;x = Mapper Ctrl

got_mir:     txa
             jsr send_byte

;we need to do a special test to check for certain 32K PRG carts

             lda #6
             sta 8000h
             lda #0
             sta 8001h		;make sure PRG is set to bank 0

             lda #4
             sta curbank	;temp for banksleft

             jsr crc_pbank	;first get CRC of 8K @ $8000

             ldx #4		;save it for comparison
savecrc:     lda temp_crc,x
             sta bank0crc,x
             dex
             bne savecrc

             lda #6		;try swapping another bank to $8000
             sta 8000h
             lda #1
             sta 8001h

             jsr crc_pbank	;get its CRC

             ldx #4		;compare it with other CRC
checkcrc:    lda temp_crc,x
             cmp bank0crc,x
             bne not32K
             dex
             bne checkcrc

is32K:       lda #128		;PRG is non-swapable 32K
             sta numpages_lo
             lda #0
             sta numpages_hi
             lda curbank
             eor #80h		;flag it so we know for later
             sta curbank
             
             jmp got_prg
             
not32K:      lda #128		;go on to check if 32K (swapable),64K, or 128K
             sta numpages_lo
             lda #0
             sta numpages_hi

             ldy #4
             jsr comp_pbank
             beq got_prg	;PRG is 32K

             lda #0
             sta numpages_lo
             lda #1
             sta numpages_hi
             asl curbank

             ldy #8
             jsr comp_pbank
             beq got_prg	;PRG is 64K

             asl numpages_hi	;PRG is 128K
             asl curbank
             
got_prg:     lda numpages_lo	;send PRG header
             jsr send_byte
             lda numpages_hi
             jsr send_byte
             lda #1
             jsr send_byte

             lda curbank
             cmp #84h		;check if special 32K
             bne continue

             lda #1		;if non-swapable 32K, need to dump all at once
             sta banksleft
             lda #0
             sta curbank
             ldx #128
             jmp next_prg

continue:    lda curbank	;if not, continue as normal
             sta banksleft
             lda #0
             sta curbank
             ldx #32

next_prg:    lda #6		;set current bank
             sta 08000h
             lda curbank
             sta 08001h
             inc curbank

             lda #0		;set read address to $8000
             sta address_lo
             lda #080h
             sta address_hi

             ldy #0		;set up for dump (x has already been set)

dump_prg:    lda (address),y
             jsr send_byte
             iny
             bne dump_prg
             inc address_hi
             dex
             bne dump_prg
             ldx #32
             dec banksleft
             bne next_prg
             
;read CHR stuff
          
             
do_chr:      lda #64
             sta numpages_lo
             lda #0
             sta numpages_hi

             ldy #16
             jsr comp_cbank
             beq got_chr	;is 16K CHR
             asl numpages_lo

             ldy #32
             jsr comp_cbank
             beq got_chr	;is 32K CHR

             lda #0
             sta numpages_lo
             lda #1
             sta numpages_hi	;is 64K CHR
             
got_chr:     lda numpages_lo
             jsr send_byte
             lda numpages_hi
             jsr send_byte
             lda #2
             jsr send_byte
             
             lda #0
             sta curbank
             lda #16
             sta banksleft

             lda numpages_lo
             cmp #64
             beq next_chr
             asl banksleft
             cmp #128
             beq next_chr
             asl banksleft
             
next_chr:    lda #2		;set current bank
             sta 08000h
             lda curbank
             sta 08001h

             lda #010h		;set up PPU to read from $1000
             jsr wr_ppu
             lda 2007h

             ldy #0		;set counters for 1K dump
             ldx #4

dump_chr:    lda 2007h
             jsr send_byte
             iny
             bne dump_chr
             dex
             bne dump_chr
             inc curbank
             dec banksleft
             bne next_chr
             
done:        lda #0
             jsr send_byte
             lda #0
             jsr send_byte
             lda #0
             jsr send_byte

             rts


;y = bank to compare
;z=same
comp_pbank:  ldx #0
             stx bank_lo	;lower pointer
             sty bank_hi	;upper pointer
             sty banksleft	;# blocks to compare

cb_loop:     lda #000h		;init pointers to $8000 & $A000
             sta address_lo
             sta uaddress_lo
             lda #080h
             sta address_hi
             lda #0a0h
             sta uaddress_hi

             lda #006h		;write to registers
             sta 08000h
             lda bank_lo
             sta 08001h
             lda #007h
             sta 08000h
             lda bank_hi
             sta 08001h

             ldx #32		;set up for 8K comp
             ldy #0

cb_loop2:    lda (address),y
             cmp (uaddress),y
             bne diff_prg
             iny
             bne cb_loop2
             inc address_hi
             inc uaddress_hi
             dex
             bne cb_loop2
             inc bank_lo
             inc bank_hi
             dec banksleft
             bne cb_loop

diff_prg:    rts

;y = bank to compare
;z=same
;for CHR
comp_cbank:  ldx #0
             stx bank_lo	;lower pointer
             sty bank_hi	;upper pointer
             sty banksleft	;# blocks to compare

cc_loop:     lda #02h		;set first bank
             sta 08000h
             lda bank_lo
             sta 08001h

             lda #03h		;set second bank
             sta 08000h
             lda bank_hi
             sta 08001h

             ldx #4
             lda #010h
             sta curbank	;reset current bank

cc_loop2:    ldy #0		;setup PPU
             lda curbank
             sta 2006h
             sty 2006h
             lda 2007h

ql:          lda 2007h		;load 256 bytes for testing
             sta 0300h,y
             iny
             bne ql

             lda curbank
             clc
             adc #4
             inc curbank
             sta 2006h       ;pointer =400h
             sty 2006h
             lda 2007h       ;garbage read

cc_loop3:    lda 2007h
             cmp 0300h,y
             bne diff_chr
             iny
             bne cc_loop3
             dex
             bne cc_loop2
             inc bank_lo
             inc bank_hi
             dec banksleft
             bne cc_loop

diff_chr:    rts

;CRC routine for 8K PRG bank @ $8000, stores crc in temp_crc

crc_pbank:   lda #0
             sta address_lo
             lda #080h
             sta address_hi	;set address to $8000

             jsr init_crc
             ldy #0
             ldx #32		;init crc for 8K of data (32 pages = 8K)

crc_ploop:   lda (address),y	;crc generation loop
             stx tx
             jsr do_crc
             ldx tx
             iny
             bne crc_ploop
             inc address_hi
             dex
             bpl crc_ploop

             jsr finish_crc	;crc done

             ldx #3
crc_psave:   lda crc0,x		;save crc in temp_crc
             sta temp_crc,x
             dex
             bpl crc_psave

             rts


             .fill 0800h-*,0ffh   ;fill rest to get 1K of data

             .end
